/**
 * Title:	Traces
 * Description: An analysis tool for traceroutes
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janb@tbm.tudelft.nl
 * @author	    Jan-Pascal van Best
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.traces;


import java.util.*;
import java.util.logging.Logger;
import java.io.*;
import java.lang.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import javax.xml.parsers.*;


public class ASHash implements Serializable {
    Hashtable[] autsystems = new Hashtable[33];
    private boolean modified;
    static final String XMLVersion = "1.0.0";
    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.traces.ASHash");

    // tables contain long integer values for IP network prefixes as keys and AutSystems as values

    public ASHash() {
        for (int i = 0; i <= 32; i++) autsystems[i] = new Hashtable();
        modified = false;
    }

    public ASHash(boolean readDefaultASes) {
        for (int i = 0; i <= 32; i++) autsystems[i] = new Hashtable();
        if (readDefaultASes) addNeededASs();
    }

    public static void writeASHash(ASHash hash, String filename) throws IOException {
        logger.info("Writing AS data to " + filename);
        PrintWriter out = new PrintWriter(new FileWriter(filename));

        out.println("<?xml version=\"1.0\"?>");
        hash.writeXML(out, "");
        out.close();
    }

    public static ASHash readASHash(String filename) throws IOException, SAXException, ParserConfigurationException {
        logger.info("Reading AS data from " + filename);
        ASHash hash = readXML(filename);

        logger.info("AS data read.");
        return hash;
    }

    public AutSystem getAS(int number) {
//        logger.info("Getting AS information for AS number "+number);
        AutSystem AS = getASFromHash(number);

        if (AS == null) {
            logger.info("Not found in hash");
            AS = new AutSystem(number);
            logger.info("got AS" + AS.number + " (" + AS.name + ")");
			if(AS.IPRanges!=null) {
				for (Iterator i = AS.IPRanges.iterator(); i.hasNext();) {
					IPRange range = (IPRange) i.next();
					byte[] netBytes = range.getNetwork().getAddress();
					Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

					autsystems[range.prefix].put(longNet, AS);
					modified = true;
				}
			}
        }
        return AS;
    }

    public AutSystem getAS(String ip) throws UnknownHostException {
        return getAS(InetAddress.getByName(ip));
    }

    public AutSystem getAS(InetAddress ip) {
        AutSystem AS = getASFromHash(ip);

        //	  logger.info("Getting AS for IP "+ip.getHostAddress());

        if (AS == null) {
            byte[] addrBytes = ip.getAddress();
            long addr = (((long) (addrBytes[0] & 255)) << 24) + (((long) (addrBytes[1] & 255)) << 16) + (((long) (addrBytes[2] & 255)) << 8) + ((long) (addrBytes[3] & 255));
            List ranges = new ArrayList();
            int number = Whois.getNumberFromIPAll(ip, ranges);
            AutSystem prevAS = getASFromHash(number);

            if (prevAS != null) {
                // Already got this one, just add ip address in hash table
                // since it didn't show up in the hash tables
                logger.info("Already have AS" + number + ", adding entry for " + ip.getHostAddress());
                autsystems[32].put(new Long(addr), prevAS);
                modified = true;
                AS = prevAS;
            } else {
                AS = new AutSystem(number);
                if (!AS.IPRanges.containsAll(ranges)) {
                    logger.info("Adding " + ranges + " to AS" + AS.number);
                    AS.IPRanges.addAll(ranges);
                }

                logger.info("got AS for ip " + ip.getHostAddress() + ": AS" + AS.number + " (" + AS.name + ")");

                boolean addressCovered = false;

                for (Iterator i = AS.IPRanges.iterator(); i.hasNext();) {
                    IPRange range = (IPRange) i.next();

                    if (!addressCovered) addressCovered = range.contains(ip);
                    byte[] netBytes = range.getNetwork().getAddress();
                    Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

                    autsystems[range.prefix].put(longNet, AS);
                    modified = true;
                    //		      logger.info("AS for prefix "+range.toString()+" added: AS"+AS.number);
                }

                if (!addressCovered) {
                    // also put requested IP address itself
                    autsystems[32].put(new Long(addr), AS);
                    modified = true;
                }
            }
        }
        return AS;
    }

    public void add(IPRange range, AutSystem AS) {
        byte[] netBytes = range.getNetwork().getAddress();
        Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

        autsystems[range.prefix].put(longNet, AS);
        modified = true;
    }

    public void addAS(AutSystem AS) {
        for (Iterator i = AS.IPRanges.iterator(); i.hasNext();) {
            IPRange range = (IPRange) i.next();

            add(range, AS);
        }
        logger.info("Added AS" + AS.number + " (" + AS.name + ")");
        modified = true;
    }

    public void addNeededASs() {
        AutSystem AS = null;

        logger.info("prefilling AS database...");

        try {
            AS = new AutSystem(2548); // DIGEX
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(2914); // VERIO
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(1239); // SPRINT
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(3561); // Cable & Wireless
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(3356); // LEVEL3
            if (AS.name.equals("")) AS.name = "ASN-LEVEL3";
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(1673);
            if (AS.name.equals("")) AS.name = "ASN-ANS-BB";
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            AS = new AutSystem(3492);
            if (AS.name.equals("")) AS.name = "ASN-ATLANTA";
        } catch (Exception e) {
            e.printStackTrace();
        };
        if (AS != null) addAS(AS);

        try {
            // 212.63.192.0-212.63.193.255 is Tele2
            AutSystem Tele2AS = new AutSystem(8873);

            addAS(Tele2AS);
            InetAddress net = InetAddress.getByName("212.63.192.0");
            byte[] netBytes = net.getAddress();
            Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

            autsystems[22].put(longNet, Tele2AS);
        } catch (Exception e) {
            e.printStackTrace();
        };

        try {
            // 192.68.191.0/24 is AS293
            AS = new AutSystem(293); // ESNET-WEST-293
            addAS(AS);
            InetAddress net = InetAddress.getByName("192.68.191.0");
            byte[] netBytes = net.getAddress();
            Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

            autsystems[24].put(longNet, AS);
        } catch (Exception e) {
            e.printStackTrace();
        };

        try {
            // 63.144.0.0/12 is AS209
            AS = new AutSystem(209); // QWEST
            addAS(AS);
            InetAddress net = InetAddress.getByName("63.144.0.0");
            byte[] netBytes = net.getAddress();
            Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

            autsystems[12].put(longNet, AS);
        } catch (Exception e) {
            e.printStackTrace();
        };

        try {
            // 192.108.148.10/32 is AS2607
            AS = new AutSystem(2607); // Slovak Technical University
            addAS(AS);
            InetAddress net = InetAddress.getByName("192.108.148.10");
            byte[] netBytes = net.getAddress();
            Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

            autsystems[32].put(longNet, AS);
        } catch (Exception e) {
            e.printStackTrace();
        };

        try {
            AS = new AutSystem(6726); // Nextra s.r.o, Bratislava
            addAS(AS);
            InetAddress net = InetAddress.getByName("192.108.148.20");
            byte[] netBytes = net.getAddress();
            Long longNet = new Long((((long) (netBytes[0] & 255)) << 24) + (((long) (netBytes[1] & 255)) << 16) + (((long) (netBytes[2] & 255)) << 8) + ((long) (netBytes[3] & 255)));

            autsystems[32].put(longNet, AS);
        } catch (Exception e) {
            e.printStackTrace();
        };

        //	  try {
        //	      AS = new AutSystem(1221);
        //	  } catch (Exception e) { e.printStackTrace(); };
        //	  if (AS != null) addAS(AS);

        modified = true;
    }

    private AutSystem getASFromHash(InetAddress ip) {
        AutSystem AS = null;

        boolean found = false;

        byte[] addrBytes = ip.getAddress();
        long addr = (((long) (addrBytes[0] & 255)) << 24) + (((long) (addrBytes[1] & 255)) << 16) + (((long) (addrBytes[2] & 255)) << 8) + ((long) (addrBytes[3] & 255));
        long net = addr;

        //	  logger.info(ip.toString());
        for (int prefix = 32; prefix > 0 && !found; net = ~(~net | (1 << (32 - prefix--)))) {
            //	      logger.info(prefix+": "+((net & (255<<24))>>24) + "." + ((net & (255<<16))>>16) + "." + ((net & (255<<8))>>8) + "." + (net & 255));
            Long Lnet = new Long(net);

            if (autsystems[prefix].containsKey(Lnet)) {
                found = true;
                AS = (AutSystem) autsystems[prefix].get(Lnet);
                //		logger.info("AS for ip "+ip.getHostAddress()+" cached: AS"+AS.number);
            }
        }

        return AS;
    }

    private AutSystem getASFromHash(int number) {
        AutSystem AS = null;

        boolean found = false;

        for (int prefix = 32; prefix > 0 && !found; prefix--) {
            for (Iterator i = autsystems[prefix].values().iterator(); i.hasNext() && !found;) {
                AS = (AutSystem) i.next();
                found = (AS.number == number);
            }
        }

        return found ? AS : null;
    }

    public void consolidate() {
        int count = 0;

        for (int prefix = 32; prefix > 0; prefix--) {
            count += autsystems[prefix].keySet().size();
        }
        logger.info("Hash contains " + count + " entries");

        count = 0;
        for (int prefix = 32; prefix > 0; prefix--) {
            int bit = 1 << (32 - prefix); // last significant bit at this prefix
            boolean doneIt;

            do {
                doneIt = false;
                for (Iterator i = autsystems[prefix].keySet().iterator(); !doneIt && i.hasNext();) {
                    Long key = (Long) i.next();
                    long laddr = key.longValue();

                    if ((laddr & bit) == 0) {
                        // we can only consolidate two prefixes if they are adjacent
                        // with only the last significant bit different
                        long laddr2 = laddr | bit;
                        AutSystem AS2 = (AutSystem) autsystems[prefix].get(new Long(laddr2));

                        if (AS2 != null) {
                            AutSystem AS1 = (AutSystem) autsystems[prefix].get(key);

                            if (AS1 == AS2) {
                                //				  String address1 = ((laddr&(255<<24))>>24) + "." + ((laddr&(255<<16))>>16) + "." + ((laddr&(255<<8))>>8) + "." + (laddr&255);
                                //				  String address2 = ((laddr2&(255<<24))>>24) + "." + ((laddr2&(255<<16))>>16) + "." + ((laddr2&(255<<8))>>8) + "." + (laddr2&255);
                                //				  logger.info("Consolidating /"+prefix+"'s "+address1+" and "+address2+" for AS"+AS1.number);
                                autsystems[prefix].remove(key);
                                autsystems[prefix].remove(new Long(laddr + bit));
                                autsystems[prefix - 1].put(key, AS1);
                                doneIt = true;
                                count++;
                            }
                        }
                    }
                }
            }
            while (doneIt);
        }
        logger.info("Consolidated " + count + " pairs");

        count = 0;
        for (int prefix = 32; prefix > 0; prefix--) {
            count += autsystems[prefix].keySet().size();
        }

        logger.info("Hash now contains " + count + " entries");
        modified = true;
    }

    public String toString() {
        String s = "Entries in hash tables:\n";

        for (int j = 0; j <= 32; j++) {
            for (Iterator i = autsystems[j].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[j].get(key);

                s = s + address + "/" + j + " -> AS" + AS.number + " (" + AS.name + ")\n";
            }
        }

        return s;
    }

    public void print() {
        logger.info("Entries in hash tables:");
        for (int j = 0; j <= 32; j++) {
            for (Iterator i = autsystems[j].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[j].get(key);

                logger.info(address + "/" + j + " -> AS" + AS.number + " (" + AS.name + ")");
            }
        }
    }

    public void printSuspiciousEntries() {
        SortedSet entries;

        entries = new TreeSet();
        logger.info("Entries in hash tables with prefix==32:\n");
        int prefix = 32;

        for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
            Long key = (Long) i.next();
            long laddr = key.longValue();
            String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
            AutSystem AS = (AutSystem) autsystems[prefix].get(key);

            entries.add(address + "/" + prefix + " -> AS" + AS.number + " (" + AS.name + ")");
        }
        for (Iterator i = entries.iterator(); i.hasNext();) {
            String s = (String) i.next();

            logger.info(s);
        }

        logger.info("ASs named \"unknown\", not AS0:\n");
        for (prefix = 0; prefix <= 32; prefix++) {
            for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[prefix].get(key);

                if (AS.number != 0 && AS.name.equals("unknown")) {
                    logger.info(address + "/" + prefix + " -> AS" + AS.number + " (" + AS.name + ")");
                }
            }
        }

        logger.info("Entries in hash tables pointing to AS0:\n");
        entries = new TreeSet();
        for (prefix = 0; prefix <= 32; prefix++) {
            for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[prefix].get(key);

                if (AS.number == 0) {
                    entries.add(address + "/" + prefix);
                }
            }
        }
        for (Iterator i = entries.iterator(); i.hasNext();) {
            String s = (String) i.next();

            logger.info(s + " -> AS0");
        }
    }

    public void removeAS0() {
        logger.info("Removing entries in hash tables pointing to AS0:\n");
        for (int prefix = 0; prefix <= 32; prefix++) {
            for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[prefix].get(key);

                if (AS.number == 0) {
                    logger.info("Removing " + address + "/" + prefix + " -> AS0");
                    i.remove();
                    modified = true;
                }
            }
        }
    }

    public void retryAS0() {
        logger.info("Retrying IP addresses pointing to AS0:\n");
        List addresses = new ArrayList(50);
        int prefix = 32;

        for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
            Long key = (Long) i.next();
            long laddr = key.longValue();
            String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
            AutSystem AS = (AutSystem) autsystems[prefix].get(key);

            if (AS.number == 0) {
                logger.info("Removing " + address + "/" + prefix + " -> AS0");
                i.remove();
                addresses.add(address);
                modified = true;
            }
        }
        for (Iterator i = addresses.iterator(); i.hasNext();) {
            String address = (String) i.next();

            try {
                AutSystem AS = getAS(address);

                logger.info("New AS for ip " + address + ": " + AS.number);
            } catch (Exception e) {
                logger.info("Failed to get AS for IP " + address);
            }
        }
    }

    public void writeXML(PrintWriter out, String indent) throws IOException {
        Set ASSet = new HashSet();

        for (int prefix = 0; prefix <= 32; prefix++) {
            ASSet.addAll(autsystems[prefix].values());
        }
        out.println(indent + "<autsystemhash version=\"" + XMLVersion + "\">");
        for (Iterator i = ASSet.iterator(); i.hasNext();) {
            AutSystem AS = (AutSystem) i.next();

            AS.writeXML(out, indent + " " + " " + " " + " ");
            out.println();
        }

        for (int prefix = 0; prefix <= 32; prefix++) {
            for (Iterator i = autsystems[prefix].keySet().iterator(); i.hasNext();) {
                Long key = (Long) i.next();
                long laddr = key.longValue();
                String address = ((laddr & (255 << 24)) >> 24) + "." + ((laddr & (255 << 16)) >> 16) + "." + ((laddr & (255 << 8)) >> 8) + "." + (laddr & 255);
                AutSystem AS = (AutSystem) autsystems[prefix].get(key);

                out.println(indent + " " + " " + " " + " " + "<hashentry range=\"" + address + "/" + prefix + "\" AS=\"" + AS.number + "\"/>");
            }
        }
        out.write(indent + "</autsystemhash>\n");
    }

    public static ASHash readXML(String filename) throws IOException, SAXException, ParserConfigurationException {
        ASHash hash = new ASHash();
        SAXParser parser = SAXParserFactory.newInstance().newSAXParser();

        parser.parse(filename, new ASHashHandler(hash));
        hash.setModified(false);
        return hash;
    }

    public void setModified(boolean modified) {
        this.modified = modified;
    }

    public boolean isModified() {
        return modified;
    }
}


class ASHashHandler extends DefaultHandler {
    ASHash hash;
    Map ASMap = new HashMap();
    AutSystem AS = null;
    StringBuffer value = new StringBuffer();

    public ASHashHandler(ASHash hash) {
        this.hash = hash;
    }

    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
        if (qName.equals("autsystemhash")) {
            String version = attributes.getValue("version");

            if (!version.equals(ASHash.XMLVersion))
                throw new SAXException("Cannot read AS hash version " + version);
        } else if (qName.equals("autsystem")) {
            int number = -1;

            try {
                number = Integer.parseInt(attributes.getValue("number"));
            } catch (NumberFormatException e) {
                throw new SAXException("Error: unparseable AS number " + attributes.getValue("number"));
            }
            AS = new AutSystem(attributes.getValue("name"), number);
        } else if (qName.equals("hashentry")) {
            IPRange range;
            int number = -1;

            try {
                number = Integer.parseInt(attributes.getValue("AS"));
            } catch (NumberFormatException e) {
                throw new SAXException("Error: unparseable AS number " + attributes.getValue("AS"));
            }
            try {
                range = new IPRange(attributes.getValue("range"));
            } catch (UnknownHostException e) {
                throw new SAXException("Cannot parse range " + attributes.getValue("range") + " of " + number);
            }
            AutSystem AS = (AutSystem) ASMap.get(new Integer(number));

            hash.add(range, AS);
        } else if (qName.equals("range")) {
            // range is handled only at endElement
        }
    }

    public void characters(char[] ch, int start, int length) throws SAXException {
        value.append(ch, start, length);
    }

    public void endElement(String uri, String localName, String qName) throws SAXException {
        //	  logger.info("End:   uri, localName, qName: "+uri+", "+localName+", "+qName);
        if (qName.equals("autsystemhash")) {
            // all is handled at beginning
        } else if (qName.equals("autsystem")) {
            ASMap.put(new Integer(AS.number), AS);
            AS = null;
        } else if (qName.equals("hashentry")) {
            // is handled at beginElement
        } else if (qName.equals("range")) {
            try {
                AS.addRange(new IPRange(value.toString().trim()));
            } catch (UnknownHostException e) {
                throw new SAXException("Cannot parse range " + value.toString().trim());
            }
            value = new StringBuffer();
        }
    }
}

