// This program is intended to query and display  from remote MySql database server
// that holds the TTM database file.

package nl.tudelft.tbm.traces;

import java.text.*;
import java.util.*;
import java.util.logging.*;

import nl.tudelft.tbm.graph.*;

public class DatesGraphFilter implements GraphFilter {
	public final static int RESOLUTION_NONE = 0;
	public final static int RESOLUTION_SECOND = 1;
	public final static int RESOLUTION_MINUTE = 2;
	public final static int RESOLUTION_HOUR = 3;
	public final static int RESOLUTION_DAY = 4;
	public final static int RESOLUTION_WEEK = 5;
	public final static int RESOLUTION_MONTH = 6;
	public final static int RESOLUTION_YEAR = 7;

	public final static String[] RESOLUTION_STRINGS={"None", "Seconds", "Minutes", "Hours",
		"Days", "Weeks", "Months", "Years"};

	private Date thresholdDate;
	private int thresholdInt;
	private boolean relative;
	private boolean enabled = true;
	private int relativeMax;
	private int resolution = RESOLUTION_NONE;

	private Date[] indices;
    protected Graph graph = null;
	private Map adaptedDatesMapNodes = null;
	private Map adaptedDatesMapEdges = null;

	private ArrayList listeners = new ArrayList();

    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.graph.EdgeCountFilter");

	public DatesGraphFilter(boolean relative, int relativeMax) {
		this.relative = relative;
		this.relativeMax = relativeMax;
    }

    public void setThreshold(Date threshold) {
		this.thresholdDate = threshold;
        notifyListeners();
    }

    public void setThreshold(int threshold) {
		this.thresholdInt = threshold;
        notifyListeners();
    }

	public Date getThresholdValue() {
		if (relative && indices!=null) {
			return indices[thresholdInt];
		} else {
			return thresholdDate;
		}
	}

	public String getThresholdString() {
		return getThresholdValue().toString();
	}

	public Date adaptToResolution(Date date) {
		switch (resolution) {
		case RESOLUTION_NONE: return date;
		case RESOLUTION_SECOND: return new Date(date.getYear(), date.getMonth(), date.getDate(), date.getHours(), date.getMinutes(), date.getSeconds());
		case RESOLUTION_MINUTE: return new Date(date.getYear(), date.getMonth(), date.getDate(), date.getHours(), date.getMinutes());
		case RESOLUTION_HOUR: return new Date(date.getYear(), date.getMonth(), date.getDate(), date.getHours(), 0);
		case RESOLUTION_DAY: return new Date(date.getYear(), date.getMonth(), date.getDate());
		case RESOLUTION_WEEK: return new Date(date.getYear(), date.getMonth(), 1+7*((date.getDate()-1)/7));
		case RESOLUTION_MONTH: return new Date(date.getYear(), date.getMonth(), 1);
		case RESOLUTION_YEAR: return new Date(date.getYear(), 0, 1);
		}
		return date;
	}

	// Returns percentile score
	protected void calculateIndices() {
		logger.info("Calculating indices");
		SortedSet dates = new TreeSet();
		adaptedDatesMapNodes = new HashMap();
		adaptedDatesMapEdges = new HashMap();
        for(Iterator i=graph.nodes.iterator(); i.hasNext(); ){
            AbstractDatedNode n = (AbstractDatedNode) i.next();
			Set adaptedDates = new HashSet();
			for(Iterator j = n.getDates().iterator(); j.hasNext(); ) {
				Date adapted = adaptToResolution((Date)j.next());
				dates.add(adapted);
				adaptedDates.add(adapted);
			}
			adaptedDatesMapNodes.put(n, adaptedDates);
		}

		for(Iterator i=graph.edges.iterator(); i.hasNext(); ){
            DelayPair e = (DelayPair) i.next();
			Set adaptedDates = new HashSet();
			for(Iterator j = e.getDates().iterator(); j.hasNext(); ) {
				Date adapted = adaptToResolution((Date)j.next());
				dates.add(adapted);
				adaptedDates.add(adapted);
			}
			adaptedDatesMapEdges.put(e, adaptedDates);
		}

		List datesList = new ArrayList(dates);
		indices = new Date[relativeMax+1];
		for(int i=0; i<relativeMax+1; i++) {
			int index = (int)Math.floor( (i*((double)1.0)/(relativeMax+1)) * datesList.size() );
			indices[i] = (Date) datesList.get(index);
		}
	}

	public boolean allowNode(Node node) {
		if (graph==null || !enabled) return true;
		if (! (node instanceof AbstractDatedNode)) return false;
		Set dates = (Set) adaptedDatesMapNodes.get(node);
		if (relative) {
			return (dates.contains(indices[thresholdInt]));
		} else {
			return (dates.contains(thresholdDate));
		}
	}

    public boolean allowEdge(Edge edge) {
		if (graph==null || !enabled) return true;
		if (! (edge instanceof DelayPair)) return false;
		Set dates = (Set) adaptedDatesMapEdges.get(edge);
		if (relative) {
			return (dates.contains(indices[thresholdInt]));
		} else {
			return (dates.contains(thresholdDate));
		}
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
		notifyListeners();
	}

	public void setResolution(int resolution) {
		this.resolution = resolution;
		if (graph != null) {
			calculateIndices();
		}
	}

	public void graphChanged(Graph graph) {
        this.graph = graph; 
		calculateIndices();
    }

    public void addListener(GraphFilterListener listener) {
		logger.info("Added listener "+listener.toString()+" to edge count filter");
        listeners.add(listener);
    }

	public void clearListeners() {
		listeners = new ArrayList();
	}

    void notifyListeners() {
        for(Iterator i=listeners.iterator(); i.hasNext();) {
            ((GraphFilterListener) (i.next())).filterChanged(this);
        }
    }
}

