/**
 * Title:		Traces
 * Copyright:	Copyright (c) Jan-Pascal van Best
 * Email:	    janpascal@vanbest.org
 * Author:	    Jan-Pascal van Best
 * Web:			http://www.tbm.tudelft.nl/webstaf/janb/software/index.html
 * API: 		http://www.tbm.tudelft.nl/webstaf/janb/software/traces/api/index.html
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.traces;

// Standard Java packages
import java.util.*;
import java.util.logging.*;
import java.io.*;
import java.lang.*;
import java.net.*;
import java.text.*;

import org.xml.sax.*;

// External packages
import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;

// Own packagages
import nl.tudelft.tbm.graph.*;
import nl.tudelft.tbm.matrix.*;

// nodes contains objects of type Object
// edges contains objects of type Pair, with x,y in nodes
// traces contains objects of type EndToEnd, with elements of hops in nodes.

/**
 * A <code>DelayGraph</code> is a {@link nl.tudelft.tbm.graph.Graph} with
 * a set of end-to-end delay measurements. These measurements are
 * 
 * A mathematical graph, consisting of a set of nodes and a set of
 * edges between them. The nodes can be of any class type, but
 * implementing the {@link Node} <code>interface</code> is preferred.
 * The edges are of type {@link Pair}. You can add nodes and edges,
 * and when you add edges the related nodes are automatically added.
 * <p>
 * Graphs can be exported to
 * <a href="http://www.cwi.nl/InfoVisu/GraphXML/index.html">GraphXML</a>,
 * <a href="http://www.caida.org/tools/visualization/otter/">Otter</a>
 * and <a href="http://www.informatik.uni-bremen.de/daVinci/">daVinci</a>
 * (<a href="http://www.davinci-presenter.de/">new site</a>) file formats.
 * The {@link #removeDullNodes} method removes nodes with exactly one
 * outgoing and one incoming edge.
 *
 * @author	    Jan-Pascal van Best
 * @version	1.0
 * @see 	Node
 * @see 	Pair
 */
public class DelayGraph extends Graph {
    Map nodeMap;
    Map edgeMap;
    Map nodesByName;
    public Set traces;

    public boolean includeTracesInXML = false;

    public final static int AS_FILTER_NONE = 0;
    public final static int AS_FILTER_BAD_IP = 1;
    public final static int AS_FILTER_UNKNOWN_AS = 2;
    public final static int AS_FILTER_LOOPS = 4;
    public final static int AS_FILTER_ALL = AS_FILTER_BAD_IP | AS_FILTER_UNKNOWN_AS | AS_FILTER_LOOPS;
    public final static String[] AS_FILTER_STRINGS = {
        "No filter", "Bad IP addresses", "Unknown AS number for IP", "Loop in trace", "All filters"};

    Matrix A = null; //
    Matrix B = null;

    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.traces");

    public DelayGraph() {
        super();
        nodeMap = new HashMap();
        edgeMap = new HashMap();
        nodesByName = new HashMap();
        traces = new HashSet();
    }

    public void addNode(Object node) {
        Object other = updateNode(node);

        if (other == node) {
            nodeMap.put(node, node);
            nodesByName.put((node instanceof Attributed) ? ((Attributed) node).getName() : node.toString(), node);
            nodes.add(node);
        }
    }

    public Object updateNode(Object node) {
        if (nodeMap.containsKey(node)) {
            Object other = nodeMap.get(node);

            if (node instanceof AbstractDatedNode && other instanceof AbstractDatedNode) {
                ((AbstractDatedNode) other).addDates(((AbstractDatedNode) node).getDates());
            }
            return other;
        } else {
            return node;
        }
    }

    public void addNodes(Collection newNodes) {
        for (Iterator i = newNodes.iterator(); i.hasNext();) {
            addNode(i.next());
        }
    }

    public Object getNodeByName(String name) {
        return nodesByName.get(name);
    }

    public void addEdge(Edge edge) {
        DelayPair other = updateEdge((DelayPair) edge);

        if (other == edge) {
            addNode(edge.x);
            addNode(edge.y);
            edgeMap.put(edge, edge);
            edges.add(edge);
        }
    }

    public DelayPair updateEdge(DelayPair edge) {
        edge.x = updateNode(edge.x);
        edge.y = updateNode(edge.y);
        if (edgeMap.containsKey(edge)) {
            DelayPair other = (DelayPair) edgeMap.get(edge);

            other.addDates(edge.getDates());
            return other;
        } else {
            return (DelayPair) edge;
        }
    }

    public void addEdge(Object from, Object to, Date date) {
        DelayPair p = new DelayPair(from, to);

        p.addDate(date);
        addEdge(p);
    }

    public void addEndToEnd(EndToEnd trace) {
        Object prev = null; // trace.from

        //	  logger.info("Adding trace: "+trace);
        for (ListIterator i = trace.hops.listIterator(); i.hasNext();) {
            Object current = i.next();

            i.set(current = updateNode(current));
            if (current instanceof AbstractDatedNode) {
                Date date = trace.getDate();
                if (date==null) {
                    logger.severe("Trace "+trace+" has null date!!!");
                } else {
                    ((AbstractDatedNode) current).addDate(trace.getDate());
                }
            }
            if (current instanceof Host) {
                if (IPRange.isBogusIPAddress(((Host) current).ip)) current = null;
            }
            if (prev != null && current != null && !current.equals(prev)) {
                addEdge(prev, current, trace.getDate());
            }
            prev = current;
        }
        traces.add(trace);
    }

    public Vector tracesContainingNode(Object node) {
        Vector result = new Vector();
        Iterator i = traces.iterator();

        while (i.hasNext()) {
            EndToEnd trace = (EndToEnd) i.next();

            for (int j = 0; j < trace.hops.size(); j++) {
                Object current = trace.hops.elementAt(j);

                if (node.equals(current)) {
                    result.add(trace);
                    break;
                }
            }
        }
        return result;
    }

    // A dull node is a node with exactly one incoming edge and
    // exactly one outgoing edge which are _not_ connected to the same node
    // (a type-1 dull node), or with two incoming and two outgoing edges
    // where the the incoming edges are connected to the same two nodes as the outgoing
    // edges (type-2 dull node), or with no incoming or outgoing edges at all
    // (type-0 dull node).

    public void removeDullNodes() {
        boolean finished = false;
        int numType0Removed = 0;
        int numType1Removed = 0;
        int numType2Removed = 0;

        while (!finished) {
            finished = true;
            //	      logger.info("Starting iteration");
            //	      logger.info("Remaining nodes: "+nodes);
            for (Iterator i = nodes.iterator(); i.hasNext();) {
                Object node = i.next();

                //		  logger.info("Looking at node: "+node);
                if ((node instanceof Node) && !((Node) node).isDeletable()) continue;
                Vector in = inEdges(node);

                if (in.size() == 0) {
                    Vector out = outEdges(node);

                    if (out.size() == 0) {
                        // we've found a type-0 dull node, remove it and relink
                        //			  logger.info("Removing type-0 node "+node);
                        Vector affectedTraces = tracesContainingNode(node);

                        removeNodeFromTraces(affectedTraces, node);
                        i.remove(); // remove current node from iterator and from nodes
                        //			      logger.info("Remaining nodes: "+nodes);
                        //			  finished = false; // not necessary, other nodes not involved
                        numType0Removed++;
                    }
                } else if (in.size() == 1) {
                    Vector out = outEdges(node);

                    if (out.size() == 1) {
                        //			  logger.info("Node has one incoming and one outgoing edge");
                        DelayPair from = (DelayPair) in.get(0);
                        DelayPair to = (DelayPair) out.get(0);

                        if (!from.x.equals(to.y)) {
                            // we've found a type-1 dull node, remove it and relink
                            //			      logger.info("Removing type-1 node "+node); // +". In: "+in+"; Out: "+out);
                            DelayPair newPair = new DelayPair(from.x, to.y);

                            newPair.addDates(from.getDates());
                            newPair.addDates(to.getDates());
                            addEdge(newPair);
                            edges.remove(from);
                            edges.remove(to);
                            Vector affectedTraces = tracesContainingNode(node);

                            removeNodeFromTraces(affectedTraces, node);
                            i.remove(); // remove current node from iterator and from nodes
                            //			      logger.info("Remaining nodes: "+nodes);
                            finished = false;
                            numType1Removed++;
                        }
                    }
                } else if (in.size() == 2) {
                    Vector out = outEdges(node);

                    if (out.size() == 2) {
                        // We've got two incoming and two outgoing edges, but are they
                        // connected to the same two nodes?
                        DelayPair edgeIn1 = (DelayPair) in.get(0);
                        DelayPair edgeIn2 = (DelayPair) in.get(1);
                        DelayPair edgeOut1 = (DelayPair) out.get(0);
                        DelayPair edgeOut2 = (DelayPair) out.get(1);

                        Object from1 = edgeIn1.x;
                        Object from2 = edgeIn2.x;
                        Object to1 = edgeOut1.y;
                        Object to2 = edgeOut2.y;

                        boolean isDull = false;

                        if (from1.equals(to2) && from2.equals(to1)) {
                            isDull = true;
                            Object temp = to1;

                            to1 = to2;
                            to2 = temp;
                        }

                        if ((from1.equals(to1) && from2.equals(to2)) || isDull) {
                            // we've found a type-2 dull node, remove it and relink
                            //			      logger.info("Removing type-2 dull node "+node); //. In: "+in+"; Out: "+out);
                            DelayPair newPair = new DelayPair(from1, to2);

                            newPair.addDates(edgeIn1.getDates());
                            if (isDull)
                                newPair.addDates(edgeOut1.getDates());
                            else
                                newPair.addDates(edgeOut2.getDates());
                            addEdge(newPair);
			
                            newPair = new DelayPair(from2, to1);
                            newPair.addDates(edgeIn2.getDates());
                            if (isDull)
                                newPair.addDates(edgeOut2.getDates());
                            else
                                newPair.addDates(edgeOut1.getDates());
                            addEdge(newPair);

                            edges.remove(edgeIn1);
                            edges.remove(edgeIn2);
                            edges.remove(edgeOut1);
                            edges.remove(edgeOut2);

                            Vector affectedTraces = tracesContainingNode(node);

                            removeNodeFromTraces(affectedTraces, node);
                            i.remove(); // remove current node from iterator and from nodes
                            //			      logger.info("Remaining nodes: "+nodes);
                            numType2Removed++;
                            finished = false;
                        }
                    }
                }
            }
        }
        logger.info("Removed dull nodes, type-0: " + numType0Removed + "; type-1: " + numType1Removed + "; type-2: " + numType2Removed +
            "; total: " + (numType0Removed + numType1Removed + numType2Removed));
    }

    public DelayGraph getSegmentGraph() {
        DelayGraph result = new DelayGraph();
        Map outHash = new HashMap();

        /* indentify crossing points: these are the nodes with
         * not exactly one incoming edges, nor exactly one outgoing edge
         */

        /* incorrect: crossing point is a node where at least two traces come in
         over the same link and leave over different links, or come in over
         different links and leave over the same link
         */

        /*
         for (Iterator i=nodes.iterator(); i.hasNext();) {
         Object o = i.next();
         Vector out = outEdges(o);
         outHash.put(o, out);
         if (out.size()<>1 || inEdges(o).size()<>1) {
         result.addNode(o);
         }
         }
         logger.info("Crossing points: "+result.nodes);

         for (Iterator i=result.nodes.iterator(); i.hasNext();) {
         Object o = i.next();
         Vector out = (Vector) outHash.get(o);
         for (Iterator j=out.iterator(); j.hasNext();) {
         DelayPair e = (DelayPair)j.next();
         Segment s = new Segment();
         s.addLink((DelayPair) e);
         DelayPair p=e;
         do {
         p = (DelayPair) ((Vector)outHash.get(p.y)).get(0);
         s.addLink(p);
         } while (!result.nodes.contains(p.y));
         result.addEdge(s);
         }
         }
         */	  return result;
    }

    public void removeUnconnectedNodes() {
        boolean finished = false;
        int numType0Removed = 0;

        while (!finished) {
            finished = true;
            for (Iterator i = nodes.iterator(); i.hasNext();) {
                Object node = i.next();
                Vector in = inEdges(node);

                if (in.size() == 0) {
                    Vector out = outEdges(node);

                    if (out.size() == 0) {
                        // we've found a type-0 dull node, remove it and relink
                        //			  logger.info("Removing type-0 node "+node);
                        Vector affectedTraces = tracesContainingNode(node);

                        removeNodeFromTraces(affectedTraces, node);
                        i.remove(); // remove current node from iterator and from nodes
                        numType0Removed++;
                    }
                }
            }
        }
        logger.info("Removed unconnected nodes (type-0): " + numType0Removed);
    }

    public void removeNodeFromTraces(Vector tr, Object node) {
        for (int j = 0; j < tr.size(); j++) {
            EndToEnd trace = (EndToEnd) tr.elementAt(j);

            for (int k = 0; k < trace.hops.size(); k++) {
                Object hop = trace.hops.elementAt(k);

                if (node.equals(hop)) {
                    trace.hops.remove(k);
                    k--;
                }
            }
        }
    }

    public DelayGraph reduceUsingAutonomousSystems(ASHash hash, int filter, boolean logTraces) {
        DelayGraph result;

        logger.info("Creating AS-reduced graph");

        result = new DelayGraph();

        // In the new graph, the nodes are of type AutSystem

        // First add nodes and fill the host.AS fields en passent
        logger.info("Translating nodes to autonomous systems...");
        int count = 1;

        for (Iterator i = nodes.iterator(); i.hasNext(); count++) {
            if (count % 72 == 0) logger.info("...");
            Object node = i.next();

            if (node instanceof Host) {
                Host host = (Host) node;
                AutSystem AS = host.AS;

                if (AS == null) {
                    if (!IPRange.isBogusIPAddress(host.ip)) {
                        AS = hash.getAS(host.ip);
                        host.AS = AS;
                    }
                }
                if (AS != null && AS.number > 0) {
                    // Mark ASs containing undeletable hosts as undeletable
                    // equivalent to: AS is deletable iff all hosts in AS are deletable
                    AS.setDeletable(AS.isDeletable() && host.isDeletable());
                    AS.addDates(host.getDates());
                    result.addNode(AS);
                }
            }
        }

        // Then recreate the traces
        logger.info("\nTranslating traces to autonomous systems...");
        count = 0;
        int numFailed = 0;
        for (Iterator i = traces.iterator(); i.hasNext(); count++) {
//            if (count % 72 == 0) logger.info("...");
            EndToEnd orig = (EndToEnd) i.next();
            EndToEnd newbie = new EndToEnd();

            newbie.setTime(orig.getTime());
            newbie.setDate(orig.getDate());

            AutSystem prev = null;
            AutSystem current = null;
            boolean first = true;
            int traceOK = 0; // This turns non-zero if there is a problem with this trace
            HashSet IPsInTrace = new HashSet();
            StringBuffer IPString = new StringBuffer();
            StringBuffer ASString = new StringBuffer();

            for (int j = 0; j < orig.hops.size(); j++) {
                Object o = orig.hops.elementAt(j);

                if (o instanceof Host) {
                    Host hop = (Host) o;
                    IPString.append(hop.ip.getHostAddress()+"\t");
                    if(IPsInTrace.contains(hop.ip)) {
                        traceOK |= AS_FILTER_LOOPS;
                    } else {
                        IPsInTrace.add(hop.ip);
                    }

                    if (hop.AS == null) {
                        if (!IPRange.isBogusIPAddress(hop.ip)) {
                            logger.info("Weird, host " + hop.getName() + " has no AS assigned, doing it now...");
                            hop.AS = hash.getAS(hop.ip);
                        } else {
                            traceOK |= AS_FILTER_BAD_IP;
                        }
                        if (hop.AS != null && hop.AS.number > 0) {
                            // Mark ASs containing undeletable hosts as undeletable
                            // equivalent to: AS is deletable iff all hosts in AS are deletable
                            hop.AS.setDeletable(hop.AS.isDeletable() && hop.isDeletable());
                            hop.AS.addDates(hop.getDates());
                            result.addNode(hop.AS);
                        }
                    }
                    current = hop.AS;
                    ASString.append(current==null?"0\t":(current.number+"\t"));
                    if(current==null || current.number<=0) {
                        traceOK |= AS_FILTER_UNKNOWN_AS;
                    }
                    if (first && current != null && current.number > 0) {
                        newbie.addHop(current);
                        first = false;
                    } else if (current != null && prev != null && prev.number > 0 && current.number > 0 && prev.number != current.number) {
                        newbie.addHop(current);
                    }
                }
                prev = current;
            }
			if (logTraces) {
				if ((traceOK&filter)==0) {
					logger.info(IPString.toString());
					logger.info(ASString.toString());
					result.addEndToEnd(newbie);
				} else {
					logger.info("rejected following trace, code "+traceOK+": "+filterToString(traceOK));
					logger.info(IPString.toString());
					logger.info(ASString.toString());
					numFailed++;
				}
			}
        }
        logger.info("Rejected "+numFailed+" of "+count+" traces.");

        // Or edges
        logger.info("\nTranslating edges to autonomous systems...");
        count = 1;
        for (Iterator i = edges.iterator(); i.hasNext(); count++) {
            if (count % 72 == 0) logger.info("...");
            DelayPair edge = (DelayPair) i.next();

            if (edge.x instanceof Host && edge.y instanceof Host) {
                AutSystem AS1 = ((Host) edge.x).AS;
                AutSystem AS2 = ((Host) edge.y).AS;

                if (AS1 != null && AS2 != null && !AS1.equals(AS2) && AS1.number > 0 && AS2.number > 0) {
                    DelayPair newEdge = new DelayPair(AS1, AS2);
                    newEdge.addDates(edge.getDates());
                    result.addEdge(newEdge);
                }
            }
        }
        logger.info("\nTranslation to autonomous systems finished");

        return result;
    }

    public static String filterToString(int filter) {
        if(filter==AS_FILTER_NONE) {
            return "none";
        } else {
            StringBuffer result = new StringBuffer();
            if ((filter & AS_FILTER_BAD_IP)!=0) result.append("Bad IP;");
            if ((filter & AS_FILTER_UNKNOWN_AS)!=0) result.append("Unknown AS for IP;");
            if ((filter & AS_FILTER_LOOPS)!=0) result.append("Trace contains loop");
            return result.toString();
        }
    }

    // routers contains an InetAddress->InetAddress mapping
    public DelayGraph reduceUsingRouters(Map ifacemap) {
        DelayGraph result = new DelayGraph();
        Map routerMap = new HashMap(); // InetAddress -> Router

        logger.info("Creating router graph");

        // First add nodes and merge to routers
        logger.info("Translating hosts to routers...");
        int count = 1;

        for (Iterator i = nodes.iterator(); i.hasNext(); count++) {
            if (count % 72 == 0) logger.info("...");
            Object node = i.next();

            if (node instanceof Host) {
                Host host = (Host) node;

                /* Possibilities re ifacemap:
                 *  - interface not in map, or points to itself
                 *  - interface in map, points to some router
                 * Possibilities re Routers (result.nodes):
                 *  - interface not known yet
                 *  - interface is itself a Router's primary interface
                 *  - interface belongs to a Router
                 */
                // First get the router

                InetAddress routerIP = (InetAddress) ifacemap.get(host.getIp());

                Router router = (Router) routerMap.get(host.getIp());

                if (router == null) {
                    router = (Router) routerMap.get(routerIP);
                    if (router == null) {
                        if (routerIP == null) {
                            router = new Router(host);
                            routerMap.put(host.getIp(), router);
                        } else {
                            router = new Router(routerIP);
                            router.mergeHost(host);
                            routerMap.put(host.getIp(), router);
                            routerMap.put(routerIP, router);
                        }
                    } else {
                        router.mergeHost(host);
                        routerMap.put(host.getIp(), router);
                    }
                } else {
                    Router router2 = (Router) routerMap.get(routerIP);

                    if (router2 == null) {
                        if (routerIP != null) {
                            router.addInterface(routerIP);
                            routerMap.put(routerIP, router);
                        }
                    } else if (router != router2) {
                        for (Iterator j = router2.interfaces.iterator(); j.hasNext();) {
                            routerMap.put(j.next(), router);
                        }
                        router.mergeRouter(router2);
                    }
                }
            }
        }
        result.addNodes(routerMap.values());

        // Then recreate the traces
        logger.info("\nTranslating traces to routers...");
        count = 1;
        for (Iterator i = traces.iterator(); i.hasNext(); count++) {
            if (count % 72 == 0) logger.info("....");
            EndToEnd orig = (EndToEnd) i.next();
            EndToEnd newbie = new EndToEnd();

            newbie.setTime(orig.getTime());
            newbie.setDate(orig.getDate());

            Router prev = null;
            Router current = null;
            boolean first = true;

            for (int j = 0; j < orig.hops.size(); j++) {
                Object o = orig.hops.elementAt(j);

                if (o instanceof Host) {
                    Host host = (Host) o;

                    current = (Router) routerMap.get(host.getIp());
                    newbie.addHop(current);
                }
                prev = current;
            }
            result.addEndToEnd(newbie);
        }

        // Or edges
        logger.info("\nTranslating edges to routers...");
        count = 1;
        for (Iterator i = edges.iterator(); i.hasNext(); count++) {
            if (count % 72 == 0) logger.info("...");
            DelayPair edge = (DelayPair) i.next();

            if (edge.x instanceof Host && edge.y instanceof Host) {
                Router router1 = (Router) routerMap.get(((Host) edge.x).getIp());
                Router router2 = (Router) routerMap.get(((Host) edge.y).getIp());
                DelayPair newEdge = new DelayPair(router1, router2);

                newEdge.addDates(edge.getDates());
                result.addEdge(newEdge);
            }
        }
        logger.info("\nTranslation to routers finished");
        return result;
    }

    public void attributeASData(ASHash hash) {
        logger.info("Attributing autonomous systems to nodes...");
        int count = 1;

        for (Iterator i = nodes.iterator(); i.hasNext(); count++) {
            Object node = i.next();

            if (node instanceof Host) {
                Host host = (Host) node;

                if (host.AS == null) {
                    if (!IPRange.isBogusIPAddress(host.ip)) {
                        host.AS = hash.getAS(host.ip);
                    }
                }
            }
        }
        logger.info("Finished attributing autonomous systems to nodes.");
    }

    public void buildEquations() {
        int numEquations = traces.size();
        int numUnknowns = edges.size();

        //	  logger.info("Graph is: " + toString());
        logger.info("Building " + numEquations + " equations with " + numUnknowns + " unknowns.");

        List links = new ArrayList(edges); // needed to give each edge a number

		for(int i=0; i<links.size(); i++) {
			Edge e = (Edge) links.get(i);
			logger.info(i+": "+e.toString());
		}

        SimpleMatrix A = new SimpleMatrix(numEquations, numUnknowns);
        SimpleMatrix B = new SimpleMatrix(numEquations, 1);

        int i = 0;

        for (Iterator it = traces.iterator(); it.hasNext(); i++) {
            EndToEnd trace = (EndToEnd) it.next();

            B.set(i, 0, trace.time);
            Object prev = null; 
            for (Iterator it2 = trace.hops.iterator(); it2.hasNext();) {
                Object current = it2.next();
                if (prev != null) {
                    DelayPair edge = new DelayPair(prev, current);
                    int j = links.indexOf(edge);

                    if (j < 0) {
                        logger.info("The link " + prev + "->" + current + " should have been identified already!");
                    } else {
                        A.set(i, j, A.get(i, j) + 1);
                    }
                }
                prev = current;
            }
        }

        logger.info("Finished building eqns");
        logger.info("A = ");
        logger.info(A.toString(8, 3));
        logger.info("B = ");
        logger.info(B.toString(8, 3));

        logger.info("Building transform");
        logger.info("Building basis for row space of A...");
        VectorSpace S = new VectorSpace(A);
        logger.info("Basis is: ");
        logger.info(S.toString(8,3));
        logger.info("Dimension of row space is "+S.getDimension());
		logger.info("Calculating nice basis...");
		SimpleMatrix Cinv = S.makeNiceBasis(A);
        logger.info("Nice basis found:");
		logger.info(Cinv.toString(8,3));

        logger.info("Transforming A...");
        SimpleMatrix ACinv = (SimpleMatrix) A.times(Cinv);
        logger.info(ACinv.toString(8,3));

        logger.info("Solving equations using new NNLS (non-negative least-square) method");
        Matrix Cx = nl.tudelft.tbm.matrix.NNLS.NNLS(ACinv, B);

        logger.info("C.x:");
        logger.info(Cx.toString(8,3));

        logger.info("Transforming back solution");
        Matrix x = Cinv.times(Cx);
        logger.info("Solution in terms of original unknowns:");
        logger.info(x.toString(8,3));

        //logger.info("Calculating direct NNLS solution for AX=B");
//        Matrix x2 = nl.tudelft.tbm.matrix.NNLS.NNLS(A, B);

        Matrix Ax = A.times(x);
//        Matrix Ax2 = A.times(x2);
        logger.info("Consistency check,");
        logger.info("   B       Ax      Ax2");
        Matrix[] vectors = new Matrix[2];
        vectors[0] = B;
        vectors[1] = Ax;
//            vectors[2] = Ax2;
        printArrays(vectors, 8, 3);

        logger.info("Calculating chisq");
        double chisq = 0;
        for(i=0; i<B.getRows(); i++) {
            double value = B.get(i,0)-Ax.get(i,0);
            chisq += value*value;
        }
        chisq /= B.getRows();
        logger.info("Chisq (sum(d^2)/N): "+chisq+"; sqrt(chisq) = "+Math.sqrt(chisq));

        logger.info("Attributing results to edges");
		for(i=0; i<numUnknowns; i++) {
			DelayPair e = (DelayPair) links.get(i);
            e.delay = x.get(i, 0);
		}

    }

    static void printArrays(Matrix[] vectors, int width, int decimals) {
        DecimalFormat format = new DecimalFormat();
        format.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.US));
        format.setMinimumIntegerDigits(1);
        format.setMaximumFractionDigits(decimals);
        format.setMinimumFractionDigits(decimals);
        format.setGroupingUsed(false);

        for(int i=0; i<vectors[0].getRows(); i++) {
            StringBuffer s = new StringBuffer();
            for (int j=0; j<vectors.length; j++) {
                String number = format.format(vectors[j].get(i,0));
                int padding = Math.max(1,width-number.length());
                s.append("                                   ".substring(0,padding));
                s.append(number);
            }
            logger.info(s.toString());
        }
    }

    static Jama.Matrix simpleToJama(SimpleMatrix A) {
        return Jama.Matrix.constructWithCopy(A.getArray());
    }

    static SimpleMatrix jamaToSimple(Jama.Matrix A) {
        return new SimpleMatrix(A.getArray());
    }

    public SimpleMatrix buildCoefficientMatrix() {
        int numEquations = traces.size();
        int numUnknowns = edges.size();

        //	  logger.info("Graph is: " + toString());
        logger.info("Building " + numEquations + " equations with " + numUnknowns + " unknowns.");

        List links = new ArrayList(edges); // needed to give each edge a number

		for(int i=0; i<links.size(); i++) {
			Edge e = (Edge) links.get(i);
			logger.info(i+": "+e.toString());
		}

        SimpleMatrix A = new SimpleMatrix(numEquations, numUnknowns);

        int i = 0;

        for (Iterator it = traces.iterator(); it.hasNext(); i++) {
            EndToEnd trace = (EndToEnd) it.next();

            Object prev = null; 
            for (Iterator it2 = trace.hops.iterator(); it2.hasNext();) {
                Object current = it2.next();
                if (prev != null) {
                    DelayPair edge = new DelayPair(prev, current);
                    int j = links.indexOf(edge);

                    if (j < 0) {
                        logger.info("The link " + prev + "->" + current + " should have been identified already!");
                    } else {
                        A.set(i, j, A.get(i, j) + 1);
                    }
                }
                prev = current;
            }
        }
        return A;
    }

    public String getDTDExtension() {
        StringBuffer s = new StringBuffer(super.getDTDExtension());

        s.append("    <!ENTITY % graphExtensions \"|path\">\n");
        s.append("    <!ELEMENT path (label,property*,vector)>\n");
        s.append("    <!ATTLIST path\n");
        s.append("        name    CDATA #REQUIRED\n");
        s.append("        delay   CDATA #IMPLIED\n");
        s.append("    >\n");
        s.append("    <!ELEMENT vector EMPTY>\n");
        s.append("    <!ATTLIST vector\n");
        s.append("        hops    CDATA #REQUIRED\n");
        s.append("    >\n");
        return s.toString();
    }
/*
    public String toXMLEdgeString(Edge edge1) {
        DelayPair edge = (DelayPair) edge1;
        StringBuffer s = new StringBuffer(4000);
        String name = edge.getName();

        if (name != null) name = name.replace('&', '+');
        String label = edge.getLabel();

        if (label != null) label = label.replace('&', '+');
        String from = (edge.x instanceof Node) ? ((Node) edge.x).getName() : edge.x.toString();
        String to = (edge.y instanceof Node) ? ((Node) edge.y).getName() : edge.y.toString();

        from = from.replace('&', '+'); // XML does not like ampersands
        to = to.replace('&', '+');
        if (name != null && !name.equals("")) {
            s.append("        <edge name=\"" + name + "\" source=\"" + from + "\" target=\"" + to + "\">\n");
        } else {
            s.append("        <edge source=\"" + from + "\" target=\"" + to + "\">\n");
        }
        if (label != null && !label.equals("")) {
            s.append("            <label>" + label + "</label>\n");
        }

        List properties = edge.getPropertyNames();

        for (Iterator i = properties.iterator(); i.hasNext();) {
            String pname = (String) i.next();
            String pvalue = edge.getProperty(pname);

            if (pvalue != null) {
                pvalue = pvalue.replace('&', '+');
                s.append("            <property name=\"" + pname + "\" content=\"" + pvalue + "\"/>\n");
            }
        }
        s.append("        </edge>\n");

        return s.toString();
    }
*/
    String toXMLEndToEndString(EndToEnd trace) {
        String s = "";
        String name = trace.getName();
        if (name != null) name = name.replace('&', '+');
//        String label = trace.getLabel();
//        if (label != null) label = label.replace('&', '+');

        s = s + "         <path name=\"" + name + "\" delay=\"" + trace.time + "\">\n";
//        if (label != null && !label.equals("")) {
//            s = s + "            <label>" + label + "</label>\n";
//        }
        List properties = trace.getPropertyNames();

        for (Iterator i = properties.iterator(); i.hasNext();) {
            String pname = (String) i.next();
            String pvalue = trace.getProperty(pname);

            if (pvalue != null) {
                pvalue = pvalue.replace('&', '+');
                s = s + "            <property name=\"" + pname + "\" content=\"" + pvalue + "\"/>\n";
            }
        }
        s = s + "            <vector hops=\"";
        for (Iterator i = trace.hops.iterator(); i.hasNext();) {
            Object hop = i.next();

            if (hop instanceof Node) {
                s = s + ((Node) hop).getName();
            } else {
                s = s + hop.toString();
            }
            if (i.hasNext()) s = s + " ";
        }
        s = s + "\"/>\n";
        s = s + "        </path>\n";

        return s;
    }

    public String getExtraXMLElements() {
        StringBuffer s = new StringBuffer(super.getExtraXMLElements());

        if (includeTracesInXML) {
            for (Iterator i = traces.iterator(); i.hasNext();) {
                EndToEnd trace = (EndToEnd) i.next();

                s.append(toXMLEndToEndString(trace) + "\n");
            }
        }
        return s.toString();
    }

    public String toString() {
        StringBuffer s = new StringBuffer();

        s.append(super.toString() + "\nTraces: ");
        for (Iterator i = traces.iterator(); i.hasNext();) {
            s.append(i.next().toString() + ",");
        }
        return s.toString();
    }

    static DelayGraph readXML(ASHash hash, String filename) throws Exception {
        logger.info("Parsing XML...");
        SAXBuilder builder = new SAXBuilder();
        Document doc = builder.build(filename);
        Element root = doc.getRootElement(); // This is a GraphXML

        Element XMLgraph = root.getChild("graph");
        DelayGraph graph = new DelayGraph();

        logger.info("Importing nodes...");
        List XMLnodes = XMLgraph.getChildren("node");

        for (Iterator i = XMLnodes.iterator(); i.hasNext();) {
            Element XMLnode = (Element) i.next();
            Node node = createNode(hash, XMLnode);

            graph.addNode(node);
        }

        logger.info("Importing edges...");
        List XMLedges = XMLgraph.getChildren("edge");
        for (Iterator i = XMLedges.iterator(); i.hasNext();) {
            Element XMLedge = (Element) i.next();
            Edge edge = createEdge(graph, XMLedge);

            graph.addEdge(edge);
        }

        logger.info("Importing paths...");
        List XMLtraces = XMLgraph.getChildren("path");
        for (Iterator i = XMLtraces.iterator(); i.hasNext();) {
			Log.logChar(".");
            Element XMLtrace = (Element) i.next();
            EndToEnd trace = createTrace(graph, XMLtrace);
            graph.addEndToEnd(trace);
        }
        return graph;
    }

    static Node createNode(ASHash hash, Element xml) {
        AbstractDatedNode node;
        String name = xml.getAttributeValue("name");

        Element XMLLabel = xml.getChild("label");
        String label = null;

        if (XMLLabel != null) label = XMLLabel.getTextNormalize();

        double x = Node.POSITION_UNKNOWN;
        double y = Node.POSITION_UNKNOWN;
        Element XMLpos = xml.getChild("position");

        if (XMLpos != null) {
            String xString = XMLpos.getAttributeValue("x");
            String yString = XMLpos.getAttributeValue("y");

            if (xString != null)
                x = Double.parseDouble(xString);
            if (yString != null)
                y = Double.parseDouble(yString);
        }

        List XMLproperties = xml.getChildren("property");
        boolean isTestBox = false;
        String dates = "";
        int ASnumber = -42;
        int testboxNumber = -42;
        double latitude = -42;
        double longitude = -42;
        String description = "";
        boolean isAutSystem = true;

        for (Iterator i = XMLproperties.iterator(); i.hasNext();) {
            Element XMLprop = (Element) i.next();
            String propName = XMLprop.getAttributeValue("name");
            String propValue = XMLprop.getAttributeValue("content");

            if (propName.equals("dates")) {
                dates = propValue;
            } else if (propName.equals("IPaddress")) {
                isAutSystem = false;
                //  Ignored: this is the name
            } else if (propName.equals("hostname")) {
                //  Ignored: this is the label
            } else if (propName.equals("ASnumber")) {
                ASnumber = Integer.parseInt(propValue);
            } else if (propName.equals("ASname")) {
                // ignored: calculated from ASnumber
            } else if (propName.equals("testboxnumber")) {
                isTestBox = true;
                testboxNumber = Integer.parseInt(propValue);
            } else if (propName.equals("latitude")) {
                latitude = Double.parseDouble(propValue);
            } else if (propName.equals("longitude")) {
                longitude = Double.parseDouble(propValue);
            } else if (propName.equals("description")) {
                description = propValue;
            }
        }

        if (isAutSystem) {
            node = hash.getAS(ASnumber);
        } else {
            if (isTestBox) {
                node = new TestBox(name, label, testboxNumber, latitude, longitude, description);
            } else {
                node = new Host(name, label);
            }
            if (ASnumber < 0)
                ((Host) node).AS = null;
            else
                ((Host) node).AS = hash.getAS(ASnumber);
        }

        node.setPosition(x, y);

        StringTokenizer st = new StringTokenizer(dates);

		node.clearDates();
        while (st.hasMoreTokens()) {
            long dateValue = Long.parseLong(st.nextToken());
            Date date = new Date(dateValue*1000);
            if (date==null) {
                logger.warning("Date of node "+name+" ("+dateValue+") give null Date");
            }
            node.addDate(date);
        }

        return node;
    }

    static Edge createEdge(DelayGraph graph, Element xml) {
        DelayPair edge;
        // The name is ignored
        String name = xml.getAttributeValue("name");
        String source = xml.getAttributeValue("source");
        String target = xml.getAttributeValue("target");
        Element XMLLabel = xml.getChild("label");
        String label = null;

        if (XMLLabel != null) label = XMLLabel.getTextNormalize();
            // The label is also ignored
        List XMLproperties = xml.getChildren("property");

        String dates = "";
        double delay = -42;
        boolean bidir = false;

        for (Iterator i = XMLproperties.iterator(); i.hasNext();) {
            Element XMLprop = (Element) i.next();
            String propName = XMLprop.getAttributeValue("name");
            String propValue = XMLprop.getAttributeValue("content");

            if (propName.equals("bidir")) {
                bidir = Boolean.valueOf(propValue).booleanValue();
            } else if (propName.equals("dates")) {
                dates = propValue;
            } else if (propName.equals("delay")) {
                delay = Double.parseDouble(propValue);
            }
        }
        Object x = graph.getNodeByName(source);
        Object y = graph.getNodeByName(target);

        if (x == null) logger.info("Error: edge source " + source + " not found!");
        if (y == null) logger.info("Error: edge target " + target + " not found!");
        edge = new DelayPair(x, y);
        edge.setBidir(bidir);
        edge.delay = delay;

        StringTokenizer st = new StringTokenizer(dates);

        while (st.hasMoreTokens()) {
            long date = Long.parseLong(st.nextToken());

            edge.addDate(new Date(date*1000));
        }

        return edge;
    }

    static EndToEnd createTrace(DelayGraph graph, Element xml) {
        EndToEnd trace;
        // The name is ignored
        String name = xml.getAttributeValue("name");
        String delayString = xml.getAttributeValue("delay");
        double delay = Double.parseDouble(delayString);

        Element XMLLabel = xml.getChild("label");
        String label = null;

        if (XMLLabel != null) label = XMLLabel.getTextNormalize();
            // The label is also ignored

        List XMLproperties = xml.getChildren("property");

        java.util.Date date = null;
        Host from = null;
        Host to = null;
		String fromString = null;
		String toString = null;
        int maxHops = Trace.MAX_HOPS_UNKNOWN;
        int packetSize = Trace.PACKET_SIZE_UNKNOWN;

//		logger.info("Reading properties...");
        for (Iterator i = XMLproperties.iterator(); i.hasNext();) {
            Element XMLprop = (Element) i.next();
            String propName = XMLprop.getAttributeValue("name");
            String propValue = XMLprop.getAttributeValue("content");

            if (propName.equals("date")) {
                date = new java.util.Date(Long.parseLong(propValue)*1000);
                if (date==null) {
                    logger.warning("Date of node "+name+" ("+propValue+") givea null Date");
                }
            } else if (propName.equals("delay")) {
                // already attribute of <path> element
            } else if (propName.equals("from")) {
				fromString = propValue;
                from = (Host) graph.getNodeByName(propValue);
            } else if (propName.equals("to")) {
				toString = propValue;
                to = (Host) graph.getNodeByName(propValue);
            } else if (propName.equals("maxhops")) {
                maxHops = Integer.parseInt(propValue);
            } else if (propName.equals("packetsize")) {
                packetSize = Integer.parseInt(propValue);
            }
        }
//		logger.info("Properties read");
//        if (from == null) logger.warning("Error: trace origin " + fromString + " not found!");
//        if (to == null) logger.warning("Error: trace destination " + toString + " not found!");

        Element XMLVector = xml.getChild("vector");
        String hopsString = XMLVector.getAttributeValue("hops");

        Vector hops = new Vector();
        StringTokenizer st = new StringTokenizer(hopsString);

        boolean isHosts = true;

//		logger.info("Getting hops from trace...");
        while (st.hasMoreTokens()) {
            String hopString = st.nextToken();
            Node hop = (Node) graph.getNodeByName(hopString);

            if (hop == null) {
//				logger.warning("Hop "+hopString+" not found in graph!");
                InetAddress ip = null;
                try {
                    ip = InetAddress.getByName(hopString);
                    if (IPRange.isBogusIPAddress(ip)) {
//						logger.info("    Creating host "+ip.getHostAddress()+"...");
                        hop = new Host(ip, hopString);
//						logger.info("    Created host");
                    }
                } catch (UnknownHostException e) {
//					logger.warning("    Unknown host!");
                }
            }
                    
            if (hop == null) {
//                logger.warning("Error: hop " + hopString + " not found, creating abstract dummy!");
                hop = new AbstractNode(hopString, hopString);
            }

            hops.add(hop);
            if (! (hop instanceof Host) ) isHosts = false;
        }
//		logger.info("Got all hops");

        if (isHosts) {
            trace = new Trace(from, to, maxHops, packetSize, hops, delay);
            trace.setDate(date);
        } else {
            trace = new EndToEnd(hops, delay, date);
        }

        return trace;
    }

    /*
     public void registerXMLHandlers() throws Exception {
     XMLSerializer.register(DelayPair.getElementName(), Class.forName("nl.tudelft.tbm.traces.DelayPair"));
     XMLSerializer.register(DelayGraph.getElementName(), Class.forName("nl.tudelft.tbm.traces.DelayGraph"));
     XMLSerializer.register("path", Class.forName("nl.tudelft.tbm.traces.Trace"));
     }

     static String getElementName() {return "graph";};

     public void writeXML(PrintWriter out, String indent) throws IOException{
     out.println(indent+"<graph isDirected=\"true\">");
     for (Iterator i=nodes.iterator(); i.hasNext(); ) {
     XMLSerializable node = (XMLSerializable) i.next();
     node.writeXML(out, indent+"    ");
     }
     for (Iterator i=edges.iterator(); i.hasNext(); ) {
     XMLSerializable edge = (XMLSerializable) i.next();
     edge.writeXML(out, indent+"    ");
     }
     if (includeTracesInXML) {
     for (Iterator i=traces.iterator(); i.hasNext(); ) {
     XMLSerializable trace = (XMLSerializable) i.next();
     trace.writeXML(out, indent+"	");
     }
     }
     out.println(indent+"</graph>");
     }

     public static XMLSerializable createObject(Attributes attributes, String contents, List subElements) throws Exception {
     DelayGraph graph = new DelayGraph();
     for (Iterator i=subElements.iterator(); i.hasNext(); ) {
     XMLSerializable o = (XMLSerializable) i.next();
     if (o instanceof Edge) {
     graph.addEdge((Edge)o);
     } else if (o instanceof EndToEnd) {
     graph.addEndToEnd((EndToEnd)o);
     } else {
     graph.addNode(o);
     }
     }
     return graph;
     }
     */
}

/*
 class NodeComparator implements Comparator {
 public int compare(Object o1, Object o2) {
 return ((Node)o1).getName().compareTo(((Node)o2).getName());
 }

 public boolean equals(Object obj) {
 return (obj==null ? false : obj.getClass() == this.getClass());
 }
 }

 class EdgeComparator implements Comparator {
 Comparator nodeComparator;
 public EdgeComparator() {nodeComparator = new NodeComparator(); }
 public EdgeComparator(Comparator nodeComparator) {this.nodeComparator = nodeComparator; }
 public int compare(Object o1, Object o2) {
 DelayPair p1 = (DelayPair) o1;
 DelayPair p2 = (DelayPair) o2;
 int result  = nodeComparator.compare(p1.x, p2.x);
 if (result==0)
 result  = nodeComparator.compare(p1.y, p2.y);
 return result;
 }
 
 public boolean equals(Object obj) {
 return (obj==null ? false : obj.getClass() == this.getClass());
 }
 }
 */


