/**
 * Title:        Traces
 * Description:  An analysis tool for traceroutes
 * Copyright:    Copyright (c) Jan-Pascal van Best
 * Email:	     janb@tbm.tudelft.nl
 * @author 	     Jan-Pascal van Best
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.traces;


import java.util.*;
import java.io.*;
import java.lang.*;
import java.net.*;
import java.text.*;

// External packages
import Jama.*;


public class NNLS {
    static double FACTOR = 0.01;
    static double TWO = 2.0;
    static double ZERO = 0.0;
    
    double A[][];   // Dimension M*N
    int MDA;
    int M;
    int N;
    double B[]; // Dimension M*1
    double X[]; // Dimension N*1
    double RNORM;
    double W[]; // Dimension N
    double ZZ[]; // Dimension M
    int INDEX[]; // N
    int MODE;
    
    int I, II, IP, ITER, ITMAX, IZ, IZ1, IZ2, IZMAX, J, JJ, JZ, L;
    int NPP1, NSETP, RTNKEY;
    double ALPHA, ASAVE, CC, DIFF, DUMMY;
    double SM, SS, T, TEMP, UNORM, UP, WMAX;
    double ZTEST;

    public NNLS(double A[][], int MDA, int M, int N, double[] B) {
        this.A = A;
        this.MDA = MDA;
        this.M = M;
        this.N = N;
        this.B = B;
        X = new double[N];
        W = new double[N];
        ZZ = new double[M];
        INDEX = new int[N];
    }

    public void calculate() {
        MODE = 1;
        if (M <= 0 || N <= 0) {
            MODE = 2;
            return;
        }

        ITER = 0;
        ITMAX = 3 * N;

        for (I = 1; I <= 20; I++) {
            X[I] = ZERO;
            INDEX[I] = I;
        }

        IZ2 = N;
        IZ1 = 1;
        NSETP = 0;
        NPP1 = 1;

        mainLoop();

        SM = ZERO;

        if (NPP1 <= M) {
            for (I = NPP1; I <= M; I++) {
                SM = SM + B[I] * B[I];
            }
        } else {
            for (J = 1; J <= N; J++) {
                W[J] = ZERO;
            }
        }
        RNORM = Math.sqrt(SM);
    }

    // Main loop in Fortran code: from "30 CONTINUE" to "GO TO 30" just before 350
    
    void mainLoop() {
        // 30
        boolean tooManyIterations = false;

        while (!(IZ1 >= IZ2 || NSETP >= M) && !tooManyIterations) {

            for (IZ = IZ1; IZ <= IZ2; IZ++) {
                J = INDEX[IZ];
                SM = ZERO;
                for (L = NPP1; L <= M; L++) SM = SM + A[L][J] * B[L];
                W[J] = SM;
            }

            boolean ready = false;

            do {
                // 60
                WMAX = ZERO;
                for (IZ = IZ1; IZ <= IZ2; IZ++) {
                    J = INDEX[IZ];
                    if (W[J] > WMAX) {
                        WMAX = W[J];
                        IZMAX = IZ;
                    }
                }

                if (WMAX <= ZERO) return;

                IZ = IZMAX;
                J = INDEX[IZ];

                ASAVE = A[NPP1][J];
                //                H12(1, NPP1, NPP1+1, M, A[1][J], 1, UP, DUMMY, 1, 1, 0); // Here A[1][J] and DUMMY are vectors!

                UNORM = ZERO;
                if (NSETP != 0) {
                    for (L = 1; L <= NSETP; L++) {
                        UNORM = UNORM + A[L][J] * A[L][J];
                    }
                }
                UNORM = Math.sqrt(UNORM);

                if (DIFF(UNORM + Math.abs(A[NPP1][J]) * FACTOR, UNORM) > ZERO) {
                    for (L = 1; L <= M; L++) ZZ[L] = B[L];
                        //            H12(2, NPP1, NPP1+1, M, A[1,J], 1, UP, ZZ, 1, 1, 1);
                    ZTEST = ZZ[NPP1] / A[NPP1][J];
                    if (ZTEST > ZERO) ready = true; // GO TO 140
                }
        
                if (!ready) {
                    A[NPP1][J] = ASAVE;
                    W[J] = ZERO;
                }
            }
            while (!ready);

            // 140
            for (L = 1; L <= M; L++) B[L] = ZZ[L];

            INDEX[IZ] = INDEX[IZ1];
            INDEX[IZ1] = J;
            IZ1++;
            NSETP = NPP1;
            NPP1++;

            if (IZ1 <= IZ2) {
                for (JZ = IZ1; JZ <= IZ2; JZ++) {
                    JJ = INDEX[JZ];
                    //                H12(2, NSETP, NPP1, M, A[1][J], 1, UP, A[1][JJ], 1, MDA, 1);
                }
            }

            if (NSETP != M) {
                for (L = NPP1; L <= M; L++)  A[L][J] = ZERO;
            }
            W[J] = ZERO;

            solveTriangular(); // RTNKEY = 1; GO TO 400

            tooManyIterations = secondaryLoop();    // 210 to 330

            if (!tooManyIterations) {
                // 330
                for (IP = 1; IP <= NSETP; IP++) {
                    I = INDEX[IP];
                    X[I] = ZZ[IP];
                }
            }
            // End of main loop
        }
    }

    // returns true on too many iterations
    boolean secondaryLoop() {
        do {
            // 210
            ITER++;
            if (ITER > ITMAX) {
                MODE = 3;
                System.out.println("NNLS quitting on iteration count");
                return true;
            }

            ALPHA = TWO;
            for (IP = 1; IP <= NSETP; IP++) {
                L = INDEX[IP];
                if (ZZ[IP] <= ZERO) {
                    T = -X[L] / (ZZ[IP] - X[L]);
                    if (ALPHA > T) {
                        ALPHA = T;
                        JJ = IP;
                    }
                }
            }

            if (ALPHA == TWO) return false;

            for (IP = 1; IP <= NSETP; IP++) {
                L = INDEX[IP];
                X[L] = X[L] + ALPHA * (ZZ[IP] - X[L]);
            }

            I = INDEX[JJ];

            boolean ready = false;

            while (!ready) {
                // 260
                X[I] = ZERO;

                if (JJ != NSETP) {
                    JJ++;

                    for (J = JJ; J <= NSETP; J++) {
                        II = INDEX[JJ];
                        INDEX[J - 1] = II;
                        //                    G1(A[J-1][II], A[J][II], CC, SS, A[J-1][II]);
                        A[J][II] = ZERO;
                        for (L = 1; L <= N; L++) {
                            if (L != II) {
                                TEMP = A[J - 1][L];
                                A[J - 1][L] = CC * TEMP + SS * A[J][L];
                                A[J][L] = -SS * TEMP + CC * A[J][L];
                            } // if
                        } // for
                        TEMP = B[J - 1];
                        B[J - 1] = CC * TEMP + SS * B[J];
                        B[J] = -SS * TEMP + CC * B[J];
                    }// for
                } // if

                NPP1 = NSETP;
                NSETP--;
                IZ1--;
                INDEX[IZ1] = I;

                ready = true;
                for (JJ = 1; JJ <= NSETP; JJ++) {
                    I = INDEX[JJ];
                    if (X[I] <= ZERO) {
                        ready = false;
                        break;
                    } // if
                } // for
            } // while

            for (I = 1; I <= M; I++) ZZ[I] = B[I];

            solveTriangular();
        }
        while (true);
    }

    void solveTriangular() {
        for (int L = 1; L <= NSETP; L++) {
            IP = NSETP + 1 - L;
            if (L != 1) {
                for (int II = 1; II <= IP; II++) ZZ[II] = ZZ[II] - A[II][JJ] * ZZ[IP + 1];
            }
            JJ = INDEX[IP];
            ZZ[IP] = ZZ[IP] / A[IP][JJ];
        }
    }

    double DIFF(double x, double y) {
        return x - y;
    }

    int H12(int MODE, int LPIVOT, int L1, int M, double U[], int IUE, int UP, double C[], int ICE, int ICV, int NCV) {
        // U[], UP, C[] are also output variables!
        // No trouble since Java passes arrays by reference
        // but we need to output UP


        return UP;
    }
}
