/**
 * Title:	 Traces
 * Description:  An analysis tool for traceroutes
 * Copyright:	 Copyright (c) Jan-Pascal van Best
 * Email:	     janb@tbm.tudelft.nl
 * @author	     Jan-Pascal van Best
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.traces;


import java.util.*;
import java.io.*;
import java.lang.*;
import java.net.*;
import java.text.*;
import org.xml.sax.*;
import nl.tudelft.tbm.graph.*;
import java.util.logging.*;


public class Trace extends EndToEnd {
    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.traces");
    public final static int TRACE_FORMAT_LINUX = 1;
    public final static int TRACE_FORMAT_WINNT = 2;

    public final static int PACKET_SIZE_UNKNOWN = -1;
    public final static int MAX_HOPS_UNKNOWN = -1;

    Host from = null;
    Host to = null;
    Host last = null;
    int maxHops = MAX_HOPS_UNKNOWN;
    int packetSize = PACKET_SIZE_UNKNOWN;

    public Trace() {
        super();
    }

    public Trace(Host from, Host to, int maxHops, int packetSize, Vector hops, double time) {
        super(hops, time);
        this.from = from;
        this.to = to;
        this.maxHops = maxHops;
        this.packetSize = packetSize;
    }

    public Trace(Trace t) {
        super(t);
        this.from = t.from;
        this.to = t.to;
        this.last = last;
        this.maxHops = t.maxHops;
        this.packetSize = t.packetSize;
    }

    public Trace(BufferedReader reader, int traceFormat) throws IOException {
        System.out.println("Trace constructor traceformat: " + traceFormat);

        switch (traceFormat) {
        case TRACE_FORMAT_LINUX:
            readLinuxHeader(reader);
            break;

        case TRACE_FORMAT_WINNT:
            readWinNTHeader(reader);
            break;
        }

        hops = new Vector();
        hops.add(from);

        time = 0.0;
        last = null;
        while (reader.ready()) {
            String s = reader.readLine();

            if (s != null && !s.equals("") && s.indexOf("Trace complete") < 0) {
                //		  System.out.println("#"+s+"#");
                //		  System.out.println("Calling Hop constructor");
                Hop hop = new Hop(s, traceFormat);

                if (hop != null) {
                    hops.addElement(hop);
                    if (hop.ms > time) time = hop.ms;
                    last = hop;
                }
            }
        }
        System.out.println("Trace time = " + time + "ms");
    }

    void readWinntHeaderOld(BufferedReader reader)  throws IOException {
        reader.readLine();
        reader.readLine();
        String s = reader.readLine();

        reader.readLine();
        int pos2 = s.indexOf("[");
        int pos1 = s.lastIndexOf(" ", pos2 - 2);
        InetAddress ip = null;

        try {
            ip = InetAddress.getByName(s.substring(pos2 + 1, s.length() - 1));
        } catch (UnknownHostException e) {
            try {
                ip = InetAddress.getByName("0.0.0.0");
            } catch (UnknownHostException f) {
            }
        }
        to = new Host(ip, s.substring(pos1 + 1, pos2 - 1));
        //	  System.out.println("To "+to.toString());
        from = Host.getLocalHost();
        //	  System.out.println("From "+from.toString());
        s = reader.readLine();
        reader.readLine();
        pos2 = s.indexOf("hops");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        maxHops = Integer.parseInt(s.substring(pos1 + 1, pos2 - 1));
        packetSize = PACKET_SIZE_UNKNOWN;
        reader.readLine();
        reader.readLine();
    }

    void readWinNTHeader(BufferedReader reader)  throws IOException {
        String s = reader.readLine(); // From: line
        int pos2 = s.indexOf("(");
        int pos1 = s.lastIndexOf(" ", pos2 - 2);

        from = new Host(s.substring(pos2 + 1, s.length() - 1), s.substring(pos1 + 1, pos2 - 1));

        s = reader.readLine(); // To name (ip): line
        pos2 = s.indexOf("(");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        to = new Host(s.substring(pos2 + 1, s.length() - 1), s.substring(pos1 + 1, pos2 - 1));

        s = reader.readLine().toLowerCase(); // Date: line
        try {
            date = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT).parse(s.substring(6, s.length()));
        } catch (ParseException e) {
            throw new IOException (e.getMessage());
        }

        reader.readLine(); // empty line
        reader.readLine(); // empty line
        s = reader.readLine(); // traceroute header line
        // format: Tracing route to <host> [<ip>]
        if (s.indexOf("Warning") >= 0) {
            s = reader.readLine(); // We had a line of type: traceroute: Warning: www.whatever.nl has multiple address; using 1.2.3.4
            //	      System.out.println("Oops, header line is really " + s);
        }

        reader.readLine(); // empty line
        s = reader.readLine(); // format: over a maximum of <n> hops
        pos2 = s.indexOf("hops");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        maxHops = Integer.parseInt(s.substring(pos1 + 1, pos2 - 1));

        packetSize = PACKET_SIZE_UNKNOWN;
    }

    void readLinuxHeader(BufferedReader reader)  throws IOException {
        String s = reader.readLine(); // From: line
        int pos2 = s.indexOf("(");
        int pos1 = s.lastIndexOf(" ", pos2 - 2);

        from = new Host(s.substring(pos2 + 1, s.length() - 1), s.substring(pos1 + 1, pos2 - 1));

        s = reader.readLine(); // To name (ip): line
        pos2 = s.indexOf("(");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        to = new Host(s.substring(pos2 + 1, s.length() - 1), s.substring(pos1 + 1, pos2 - 1));

        s = reader.readLine().toLowerCase(); // Date: line
        try {
            date = DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT).parse(s.substring(6, s.length()));
        } catch (ParseException e) {
            throw new IOException (e.getMessage());
        }

        s = reader.readLine(); // traceroute header line
        // format: traceroute to <host> (<ip>), <n> hops max, <b> byte packets
        if (s.indexOf("Warning") >= 0) {
            s = reader.readLine(); // We had a line of type: traceroute: Warning: www.whatever.nl has multiple address; using 1.2.3.4
            //	      System.out.println("Oops, header line is really " + s);
        }
        pos2 = s.indexOf("hops");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        maxHops = Integer.parseInt(s.substring(pos1 + 1, pos2 - 1));

        pos2 = s.indexOf("byte");
        pos1 = s.lastIndexOf(" ", pos2 - 2);
        packetSize = Integer.parseInt(s.substring(pos1 + 1, pos2 - 1));
    }

    public String toString() {
        String s = "Trace from " + from.toString() + " to " + to.toString() + " at " + date.toString() + " in " + time + " ms, max " + maxHops + " hops, " + packetSize + " bytes.\n";

        for (Enumeration e = hops.elements(); e.hasMoreElements();) {
            s = s + ((Host) e.nextElement()).toString() + ",";
        }
        return s;
    }

    public String getName() {
        return ((from == null) ? "-" : from.getName()) +
            "to" +
            ((to == null) ? "-" : to.getName()) +
            "@" + Integer.toHexString(hashCode());
    }

    public String getLabel() {
        return ((from == null) ? "-" : from.getName()) +
            " to " +
            ((to == null) ? "-" : to.getName()) +
            ((date == null) ? "" : " at " + date.toString());
    }

    public List getPropertyNames() {
        String[] strings = {"from", "to", "maxhops", "packetsize"};
        List result = new ArrayList(Arrays.asList(strings));

        result.addAll(super.getPropertyNames());
        return result;
    }

    public String getProperty(String name) {
        String result;

        if (name.equals("from")) {
            result = (from == null) ? "" : from.getName();
        } else if (name.equals("to")) {
            result = (to == null) ? "" : to.getName();
        } else if (name.equals("maxhops")) {
            result = "" + maxHops;
        } else if (name.equals("packetsize")) {
            result = "" + packetSize;
        } else
            result = super.getProperty(name);
        return result;
    }

    /*
     static String getElementName() {return "path";};

     public void writeXML(PrintWriter out, String indent) throws IOException{
     String name = getName().replace('&','+');
     String label = getLabel();
     if (label!=null) label=label.replace('&','+');
     out.println(indent+"<path name=\""+name+"\" delay=\""+time+"\">");
     if (label != null && !label.equals("")) {
     out.println(indent+"    <label>"+label+"</label>");
     }
     List properties = getPropertyNames();
     for (Iterator i=properties.iterator(); i.hasNext(); ) {
     String pname = (String) i.next();
     String pvalue = getProperty(pname);
     if (pvalue!=null) {
     pvalue = pvalue.replace('&','+');
     out.println(indent+"	<property name=\""+pname+"\" content=\""+pvalue+"\"/>");
     }
     }
     out.print(indent+"    <vector hops=\"");
     for (Iterator i=hops.iterator(); i.hasNext(); ) {
     Object hop = i.next();
     if (hop instanceof Node) {
     out.print(((Node) hop).getName());
     } else {
     out.print(hop.toString());
     }
     if (i.hasNext()) out.print(" ");
     }
     out.println("\"/>");
     out.println(indent+"</path>");
     }

     public static XMLSerializable createObject(Attributes attributes, String contents, List subElements) throws Exception {
     logger.info("Creating new Trace from XML");
     String name = attributes.getValue("name");
     String delayString = attributes.getValue("delay");
     double delay = Double.parseDouble(delayString);
     // FIXME: handle <label>, <property> and <vector>
    Trace result = new Trace();
    result.name = name;
    result.time = delay;
    //	  System.out.println("	  Creating edge: "+x.toString()+"->"+y.toString());
    return result;
    }
    */

}

