/**
 * Title:	 Traces
 * Description:  An analysis tool for traceroutes
 * Copyright:	 Copyright (c) Jan-Pascal van Best
 * Email:	 janb@tbm.tudelft.nl
 * Author:	 Jan-Pascal van Best
 **************************************************************************
 * Sources: RIPE NCC Database Group, Whois queries in RIPE database v3.0,
 *	    RIPE document NewDB-1, December 27, 2000
 *	    C. Alaettinoglu et.al., Routing Policy Specification Language
 *	    (RPSL), RFC2622, June 1999
 *	    A.M.R. Magee, RIPE Database Documentation, RIPE document ripe-157,
 *	    29 May, 1997
 *	    whoisRIP-1.0 source code
 *	    List of whois servers at ftp://ftp.ra.net/radb/list.html
 *
 **************************************************************************
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package nl.tudelft.tbm.traces;


import java.util.*;
import java.io.*;
import java.lang.*;
import java.net.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.xml.sax.*;
import java.util.logging.*;

public class Whois {
    private static Logger logger = Logger.getLogger("nl.tudelft.tbm.traces");
    protected final static String CRLF = "\r\n";
    protected final static String WHOIS_VERSION = "-V";

    protected final static Whois[] DEFAULT_SERVERS = {
            //	  address, port, isRIP-type server, supports inverse lookups
            new Whois("whois.ripe.net", 43, true, true),
            new Whois("rr.arin.net", 43, false, true),
            new Whois("whois.apnic.net", 43, true, true),
            new Whois("whois.internet2.edu", 43, false, false),
            //	  new Whois("rs.internic.net", 43, false), // Only domain name lookups
            new Whois("whois.radb.net", 43, false, false),
            new Whois("whois.koren21.net", 43, false, false),
            new Whois("whois.connect.com.au", 43, false, true),
            new Whois("whois.ra.net", 43, false, false),
            new Whois("whois.ntt.net", 43, false, false),
            new Whois("whois.cw.net", 43, true, true),
            new Whois("rr.verio.net", 43, false, false),
            new Whois("rr.level3.net", 43, false, true),
            new Whois("whois.altdb.net", 43, false, false),
            new Whois("irr.iij.ad.jp", 43, false, false),
            //	  new Whois("whois.gctr.net", 43, false), // Seems to be down
            new Whois("rr.sinet.ad.jp", 43, false, false),
            new Whois("whois.semaphore.net", 43, false, false),
            new Whois("whois.scc.net", 43, false, false), // This one gives /1 and /2 routes!!
            new Whois("whois.look.ca", 43, false, false),
            new Whois("whois.nic.mil", 43, false),
            new Whois("irr.nic.mil", 43, false, false)
            //	  new Whois("whois.nic.gov", 43, false) // Only domains
        };
    // There should also be servers at: mci.net; isi.edu; garr.it; ans.net
    // (see whoisRIP sources)

    protected static List servers = new ArrayList(0);

    String host;
    int port;
    boolean RIP;
    boolean doInverse;

    public Whois (String host, int port, boolean isRIP, boolean supportsInverse) {
        this.host = host;
        this.port = port;
        this.RIP = isRIP;
        this.doInverse = supportsInverse;
    }

    public Whois (String host, int port, boolean isRIP) {
        this.host = host;
        this.port = port;
        this.RIP = isRIP;
        this.doInverse = true;
    }

    public int getNumberFromIP(InetAddress ip, List ranges) {
        int number = 0;

        try {
            //	      logger.info("Connecting to "+host);
            Socket socket = new Socket(host, port);
            BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            BufferedWriter out = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));

            String query;

            if (RIP)
                query = ip.getHostAddress();
            else
                query = ip.getHostAddress();

                //	      logger.info("Query: "+query);
            out.write(query + CRLF);
            out.flush();

            boolean routeObject = false;
            boolean rtrObject = false;
            boolean prevEmpty = true;
            boolean finished = false;

            IPRange range = null;

            while (!finished) {
                String line = in.readLine();

                finished = (line == null);
                if (!finished) {
                    //		      logger.info(prevEmpty+" "+routeObject+" "+line);
                    if (prevEmpty && line.startsWith("route:")) {
                        routeObject = true;
                        //			  logger.info(host+": "+line.substring(7).trim());
                        range = new IPRange(line.substring(7).trim());
                        if (range.prefix < 8) {
                            range = null;
                            routeObject = false;
                        }
                    } else if (prevEmpty && line.startsWith("*rt:")) {
                        routeObject = true;
                        //			  logger.info(host+": "+line.substring(5).trim());
                        range = new IPRange(line.substring(5).trim());
                        if (range.prefix < 8) {
                            range = null;
                            routeObject = false;
                        }
                    }

                    if (prevEmpty && (line.startsWith("inet-rtr:") || line.startsWith("*ir:"))) {
                        rtrObject = true;
                        //			  logger.info(line);
                    }

                    if (routeObject && line.startsWith("origin:")) {
                        number = Integer.parseInt(line.substring(7).trim().substring(2));
                        //			  logger.info(number);
                        if (range != null) ranges.add(range);
                        finished = true;
                    } else if (routeObject && line.startsWith("*or:")) {
                        number = Integer.parseInt(line.substring(4).trim().substring(2));
                        if (range != null) ranges.add(range);
                        finished = true;
                    } else if (rtrObject && line.startsWith("local-as:")) {
                        //			  logger.info(line);
                        number = Integer.parseInt(line.substring(9).trim().substring(2));
                        finished = true;
                    } else if (rtrObject && line.startsWith("*la:")) {
                        //			  logger.info(line);
                        number = Integer.parseInt(line.substring(4).trim().substring(2));
                        finished = true;
                    }
                    if (line.equals("")) {
                        if (number > 0) finished = true;
                        prevEmpty = true;
                        routeObject = false;
                        rtrObject = false;
                        range = null;
                    } else prevEmpty = false;
                }
            }
            socket.close();
        } catch (IOException e) {
            //number = 0;
        }
        return number;
    }

    public String getNameFromNumber(int number) {
        String name = "";

        try {
            Socket socket = new Socket(host, port);
            BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            BufferedWriter out = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));

            String query;

            if (RIP)
                query = "AS" + number;
            else
                query = "AS" + number;

            out.write(query + CRLF);
            out.flush();

            boolean finished = false;
            boolean autnumObject = false;
            boolean prevEmpty = true;

            while (!finished) {
                String line = in.readLine();

                finished = (line == null);
                if (!finished) {
                    //		      logger.info(line);
                    if (prevEmpty && (line.startsWith("aut-num:") || line.startsWith("*an:"))) autnumObject = true;
                    if (autnumObject && line.startsWith("descr:")) {
                        name = line.substring(6).trim();
                        finished = true;
                    } else if (autnumObject && line.startsWith("*de:")) {
                        name = line.substring(4).trim();
                        finished = true;
                    }
                    if (line.equals("")) {
                        prevEmpty = true;
                        autnumObject = false;
                    } else prevEmpty = false;
                }
            }
            //	      if (name.equals("")) name = "AS"+number;
            socket.close();
        } catch (IOException e) {
            // name = "";
        }
        return name;
    }

    public List getRangesFromNumber(int number) {
        List ranges = new ArrayList();

        try {
            Socket socket = new Socket(host, port);
            BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            BufferedWriter out = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));

            String query;

            if (RIP) {
                query = "-Trt -ior AS" + number;
            } else {
                query = "-T route -i aut-num AS" + number;
            }
            //	      logger.info("Query: "+query);
            out.write(query + CRLF);
            out.flush();

            boolean finished = false;

            while (!finished) {
                String line = in.readLine();

                finished = (line == null);
                if (!finished) {
                    //		      logger.info(line);
                    if (line.startsWith("route:")) {
                        ranges.add(new IPRange(line.substring(7).trim()));
                    } else if (line.startsWith("*rt:")) {
                        ranges.add(new IPRange(line.substring(5).trim()));
                    }
                }
            }
            socket.close();
        } catch (IOException e) {
        }
        return ranges;
    }

    public int getNumberFromIP(InetAddress ip) {
        List ranges = new ArrayList(1);

        return getNumberFromIP(ip, ranges);
    }

    public static int getNumberFromIPAll(InetAddress ip) {
        List ranges = new ArrayList(1);

        return getNumberFromIPAll(ip, ranges);
    }

    public static int getNumberFromIPAll(InetAddress ip, List ranges) {
        int number = 0;

        for (Iterator i = servers.iterator(); i.hasNext();) {
            //	      logger.info("Asking AS number for IP "+ip.toString()+" to "+DEFAULT_SERVERS[i].host);
            Whois server = (Whois) i.next();
            number = server.getNumberFromIP(ip, ranges);
            if (number > 0) break;
        }
        return number;
    }

    public static String getNameFromNumberAll(int number) {
        String name = "";

        for (Iterator i = servers.iterator(); i.hasNext();) {
            //	      logger.info("Asking AS name for AS"+number+" to "+DEFAULT_SERVERS[i].host);
            Whois server = (Whois) i.next();
            name = server.getNameFromNumber(number);
            if (!name.equals("")) break;
        }
        if (name.equals("")) {
            name = "unknown";
        }
        return name;
    }

    public static List getRangesFromNumberAll(int number) {
        List IPRanges = null;

        for (Iterator i = servers.iterator(); i.hasNext();) {
            Whois server = (Whois) i.next();
            if (server.doInverse) {
                //		  logger.info("Asking ranges for AS"+number+" to "+DEFAULT_SERVERS[i].host);
                IPRanges = server.getRangesFromNumber(number);
                if (IPRanges.size() != 0) break;
            }
        }
        return IPRanges;
    }

    public static void test1(String host, boolean isRIP, int number) {
        List ranges = null;
        Whois server = new Whois(host, 43, isRIP);

        //	  logger.info("*****************");
        //	  logger.info(server.host);
        ranges = server.getRangesFromNumber(number);
        //	  logger.info(ranges.toString());
    }

    public static void test() throws IOException {
        // known
        //	  test1("whois.ripe.net", true, 7176);
        //	  test1("rr.arin.net", false, 7176);
        //	  test1("whois.cw.net", true, 701);
        //	  test1("rr.level3.net", false, 3967);
        //	  test1("whois.connect.com.au", false, 2764);

        // No support for reverse lookups
        /*
         test1("whois.internet2.edu", true, 7176);
         test1("whois.radb.net", true, 7176);
         test1("whois.koren21.net", true, 7176);
         test1("whois.ra.net", true, 7176);
         test1("whois.ntt.net", true, 7176);
         test1("rr.verio.net", true, 7176);
         test1("irr.iij.ad.jp", true, 7176);
         test1("whois.semaphore.net", true, 7176);
         test1("rr.sinet.ad.jp", true, 7176);
         test1("whois.scc.net", true, 7176);
         test1("whois.look.ca", true, 7176);
         test1("irr.nic.mil", false, 7176);
         test1("whois.altdb.net", false, 7176);
         */

        // unknown
        test1("whois.apnic.net", true, 2764);
        //	  test1("whois.gctr.net", true, 7176); Out to lunch
        test1("whois.nic.mil", true, 5927);
        //	  logger.info("Finished");
    }


    public static void useDefaultServers() {
        servers = new ArrayList();
        for (int i = 0; i < DEFAULT_SERVERS.length; i++) {
            servers.add(DEFAULT_SERVERS[i]);
        }
    }

    String toXMLString(String indent) {
        return indent + "<whois address=\""+host+"\" port=\""+
            port+"\" RIP=\""+RIP+"\" supportsInverse=\""+doInverse+"\"/>";
    }

    public static void exportServersXML(String filename) throws IOException {
        //	  logger.info("Exporting GraphXML to "+filename);
        FileWriter f = new FileWriter(filename);
        PrintWriter p = new PrintWriter(f);

        p.println("<?xml version=\"1.0\"?>");
        p.println("<whoisservers>");

        for (Iterator i = servers.iterator(); i.hasNext();) {
            Whois server = (Whois) i.next();
            p.println(server.toXMLString("    "));
        }
        p.println("</whoisservers>");
        p.flush();
        p.close();
        f.close();
    }

    private static Whois createWhois(Element xml) {
        String address;
        int port;
        boolean RIP;
        boolean doInverse;

        address = xml.getAttributeValue("address");
        try {
            port = Integer.parseInt(xml.getAttributeValue("port"));
        } catch (NumberFormatException e) {
            port = 43;
        }
        RIP = Boolean.valueOf(xml.getAttributeValue("RIP")).booleanValue();
        doInverse = Boolean.valueOf(xml.getAttributeValue("supportsInverse")).booleanValue();

        return new Whois(address, port, RIP, doInverse);
    }

    public static void readServersXML(String filename) throws Exception {
        servers = new ArrayList();
        SAXBuilder builder = new SAXBuilder();
        Document doc = builder.build(filename);

        Element root = doc.getRootElement(); // This is a <whoisservers>

        List XMLservers = root.getChildren("whois");
        for (Iterator i = XMLservers.iterator(); i.hasNext();) {
            Element XMLwhois = (Element) i.next();
            Whois whois = createWhois(XMLwhois);
            servers.add(whois);
        }
    }

}

